import logging
import rlp
from datetime import datetime, timedelta
from time import sleep

from trezorlib.tools import parse_path
from trezorlib import ethereum
from web3 import Web3, exceptions

from axie_utils.abis import AXIE_ABI
from axie_utils.utils import (
    get_nonce,
    RONIN_PROVIDER,
    AXIE_CONTRACT,
    TIMEOUT_MINS,
    USER_AGENT
)


class Transfer:
    def __init__(self, from_acc, from_private, to_acc, axie_id):
        self.w3 = Web3(
            Web3.HTTPProvider(
                RONIN_PROVIDER,
                request_kwargs={"headers": {"content-type": "application/json", "user-agent": USER_AGENT}}))
        self.from_acc = from_acc.replace("ronin:", "0x")
        self.from_private = from_private
        self.to_acc = to_acc.replace("ronin:", "0x")
        self.axie_id = axie_id

    def execute(self):
        # Load ABI
        axie_contract = self.w3.eth.contract(
            address=Web3.toChecksumAddress(AXIE_CONTRACT),
            abi=AXIE_ABI
        )
        # Get Nonce
        nonce = get_nonce(self.from_acc)
        # Build transaction
        transaction = axie_contract.functions.safeTransferFrom(
            Web3.toChecksumAddress(self.from_acc),
            Web3.toChecksumAddress(self.to_acc),
            self.axie_id
        ).buildTransaction({
            "chainId": 2020,
            "gas": 492874,
            "from": Web3.toChecksumAddress(self.from_acc),
            "gasPrice": self.w3.toWei("1", "gwei"),
            "value": 0,
            "nonce": nonce
        })
        # Sign Transaction
        signed = self.w3.eth.account.sign_transaction(
            transaction,
            private_key=self.from_private
        )
        # Send raw transaction
        self.w3.eth.send_raw_transaction(signed.rawTransaction)
        # get transaction _hash
        _hash = self.w3.toHex(self.w3.keccak(signed.rawTransaction))
        # Wait for transaction to finish or timeout
        start_time = datetime.now()
        while True:
            # We will wait for max 5min for this trasnfer to happen
            if datetime.now() - start_time > timedelta(minutes=TIMEOUT_MINS):
                success = False
                logging.info(f"Important: Transfer {self}, timed out!")
                break
            try:
                recepit = self.w3.eth.get_transaction_receipt(_hash)
                if recepit["status"] == 1:
                    success = True
                else:
                    success = False
                break
            except exceptions.TransactionNotFound:
                logging.info(f"Waiting for transfer '{self}' to finish (Nonce:{nonce})...")
                # Sleep 10 seconds not to constantly send requests!
                sleep(10)
        if success:
            logging.info(f"{self} completed! Hash: {_hash} - "
                         f"Explorer: https://explorer.roninchain.com/tx/{str(_hash)}")
            return _hash
        else:
            logging.info(f"{self} failed")
            return

    def __str__(self):
        return (f"Axie Transfer of axie ({self.axie_id}) from account ({self.from_acc.replace('0x', 'ronin:')}) "
                f"to account ({self.to_acc.replace('0x', 'ronin:')})")


class TrezorTransfer:
    def __init__(self, from_acc, client, bip_path, to_acc, axie_id):
        self.w3 = Web3(
            Web3.HTTPProvider(
                RONIN_PROVIDER,
                request_kwargs={"headers": {"content-type": "application/json", "user-agent": USER_AGENT}}))
        self.from_acc = from_acc.replace("ronin:", "0x")
        self.to_acc = to_acc.replace("ronin:", "0x")
        self.axie_id = axie_id
        self.client = client
        self.bip_path = parse_path(bip_path)
        self.gwei = self.w3.toWei('1', 'gwei')
        self.gas = 250000

    def execute(self):
        axie_contract = self.w3.eth.contract(
            address=Web3.toChecksumAddress(AXIE_CONTRACT),
            abi=AXIE_ABI
        )
        # Get Nonce
        nonce = get_nonce(self.from_acc)
        # Build transaction
        transfer_tx = axie_contract.functions.safeTransferFrom(
            Web3.toChecksumAddress(self.from_acc),
            Web3.toChecksumAddress(self.to_acc),
            self.axie_id
        ).buildTransaction({
            "chainId": 2020,
            "gas": self.gas,
            "from": Web3.toChecksumAddress(self.from_acc),
            "gasPrice": self.w3.toWei("1", "gwei"),
            "value": 0,
            "nonce": nonce
        })
        data = self.w3.toBytes(hexstr=transfer_tx['data'])
        to = self.w3.toBytes(hexstr=AXIE_CONTRACT)
        sig = ethereum.sign_tx(
            self.client,
            n=self.bip_path,
            nonce=nonce,
            gas_price=self.gwei,
            gas_limit=self.gas,
            to=AXIE_CONTRACT,
            value=0,
            data=data,
            chain_id=2020
        )
        transaction = rlp.encode((nonce, self.gwei, self.gas, to, 0, data) + sig)
        # Send raw transaction
        self.w3.eth.send_raw_transaction(transaction)
        # Get transaction _hash
        _hash = self.w3.toHex(self.w3.keccak(transaction))
        # Wait for transaction to finish or timeout
        start_time = datetime.now()
        while True:
            # We will wait for max 5min for this trasnfer to happen
            if datetime.now() - start_time > timedelta(minutes=TIMEOUT_MINS):
                success = False
                logging.info(f"Transfer {self}, timed out!")
                break
            try:
                recepit = self.w3.eth.get_transaction_receipt(_hash)
                if recepit["status"] == 1:
                    success = True
                else:
                    success = False
                break
            except exceptions.TransactionNotFound:
                logging.info(f"Waiting for transfer '{self}' to finish (Nonce:{nonce})...")
                # Sleep 10 seconds not to constantly send requests!
                sleep(10)
        if success:
            logging.info(f"{self} completed! Hash: {_hash} - "
                         f"Explorer: https://explorer.roninchain.com/tx/{str(_hash)}")
            return _hash
        else:
            logging.info(f"{self} failed")
            return

    def __str__(self):
        return (f"Axie Transfer of axie ({self.axie_id}) from account ({self.from_acc.replace('0x', 'ronin:')}) "
                f"to account ({self.to_acc.replace('0x', 'ronin:')})")



import rlp
import logging
from datetime import datetime, timedelta
from time import sleep

from trezorlib import ethereum
from trezorlib.tools import parse_path
from web3 import Web3, exceptions

from axie_utils.abis import SCATTER_ABI, APPROVE_ABI
from axie_utils.utils import (
    get_nonce,
    check_balance,
    SCATTER_CONTRACT,
    TOKEN,
    RONIN_PROVIDER,
    USER_AGENT,
    TIMEOUT_MINS
)
    

class Scatter:
    def __init__(self, token, from_acc, from_private, to_ronin_ammount_dict):
        self.w3 = Web3(
            Web3.HTTPProvider(
                RONIN_PROVIDER,
                request_kwargs={"headers": {"content-type": "application/json", "user-agent": USER_AGENT}}))
        self.token = token.lower()
        if self.token != 'ron':
            self.token_contract = self.w3.eth.contract(
                address=Web3.toChecksumAddress(TOKEN[self.token]),
                abi=APPROVE_ABI
            )
        self.from_acc = from_acc.replace("ronin:", "0x")
        self.from_private = from_private
        self.contract = self.w3.eth.contract(
            address=Web3.toChecksumAddress(SCATTER_CONTRACT),
            abi=SCATTER_ABI
        )
        self.to_list = []
        self.amounts_list = []
        for k,v in to_ronin_ammount_dict.items():
            self.to_list.append(k.replace("ronin:", "0x"))
            self.amounts_list.append(v)
   
    def is_contract_accepted(self):
        allowance = self.token_contract.functions.allowance(
            Web3.toChecksumAddress(self.from_acc),
            Web3.toChecksumAddress(SCATTER_CONTRACT)).call()
        if int(allowance) > sum(self.amounts_list):
            return True
        return self.approve_contract()

    def approve_contract(self):
        approve_tx = self.token_contract.functions.approve(
            Web3.toChecksumAddress(SCATTER_CONTRACT),
            115792089237316195423570985008687907853269984665640564039457584007913129639935
        ).buildTransaction({
            "gas": 1000000,
            "gasPrice": self.w3.toWei(1, "gwei"),
            "nonce": get_nonce(self.from_acc)
        })
        signed_approval = self.w3.eth.account.sign_transaction(
            approve_tx,
            private_key=self.from_private
        )
        self.w3.eth.send_raw_transaction(signed_approval.rawTransaction)
        approve_hash = self.w3.toHex(self.w3.keccak(signed_approval.rawTransaction))
        approved = self.w3.eth.wait_for_transaction_receipt(approve_hash, timeout=240)
        if approved['status'] == 1:
            return True
        return False

    def increase_gas_tx(self, nonce):
        # check nonce is still available, do nothing if nonce is not available anymore
        if nonce != get_nonce(self.from_acc):
            return
        # Increase gas price to get tx unstuck
        return self.execute(1.01, nonce)

    def execute_token(self, gas_price=1, nonce=None):
        # Check token is approved
        if not self.is_contract_accepted():
            logging.warning(f"Token {self.token} is not approved to use scatter, "
                            "you can re-try or manually accept it on "
                            "scatter website (https://scatter.roninchain.com/).")
            return

        # Check enough balance is present
        if not check_balance(self.from_acc, self.token) >= sum(self.amounts_list) and check_balance(self.from_acc, 'ron') >= 0.011:
            logging.warning("Not enough {TOKEN[self.token]} balance or not enough RON to pay for the tx")
            return
        
        # Get Nonce
        if nonce is None:
            nonce = get_nonce(self.from_acc)
        # Build transaction
        transaction = self.contract.functions.disperseTokenSimple(
            TOKEN[self.token],
            self.to_list,
            self.amounts_list
        ).buildTransaction({
            "chainId": 2020,
            "gas": 1000000,
            "gasPrice": self.w3.toWei(str(gas_price), "gwei"),
            "nonce": nonce
        })
        # Sign Transaction
        signed = self.w3.eth.account.sign_transaction(
            transaction,
            private_key=self.from_private
        )
        # Send raw transaction
        self.w3.eth.send_raw_transaction(signed.rawTransaction)
        # get transaction _hash
        _hash = self.w3.toHex(self.w3.keccak(signed.rawTransaction))
        # Wait for transaction to finish or timeout
        start_time = datetime.now()
        while True:
            # We will wait for max 5minutes for this tx to respond, if it does not, we will re-try
            if datetime.now() - start_time > timedelta(minutes=TIMEOUT_MINS):
                success = False
                logging.info(f"Transaction {self}, timed out!")
                break
            try:
                recepit = self.w3.eth.get_transaction_receipt(_hash)
                if recepit["status"] == 1:
                    success = True
                else:
                    success = False
                break
            except exceptions.TransactionNotFound:
                # Sleep 10s while waiting
                sleep(10)
                logging.info(f"Waiting for transaction '{self}' to finish (Nonce:{nonce})...")

        if success:
            logging.info(f"Transaction {self} completed! hash: {_hash} - "
                        f"Explorer: https://explorer.roninchain.com/tx/{str(_hash)}")
            return _hash
        else:
            logging.info(f"Transaction {self} failed. Trying to augment gas price to unstuck it.")
            self.increase_gas_tx(nonce)
    
    
    def execute_ron(self, gas_price=1, nonce=None):
        # Check enough balance is present
        if not check_balance(self.from_acc, 'ron') >= (sum(self.amounts_list) + 0.011):
            logging.warning("Not enough RON balance to scatter and pay the tx.")
            return
                
        # Get Nonce
        if nonce is None:
            nonce = get_nonce(self.from_acc)
        # Build transaction
        transaction = self.contract.functions.disperseEther(
            self.to_list,
            self.amounts_list
        ).buildTransaction({
            "chainId": 2020,
            "gas": 1000000,
            "gasPrice": self.w3.toWei(str(gas_price), "gwei"),
            "nonce": nonce
        })
        # Sign Transaction
        signed = self.w3.eth.account.sign_transaction(
            transaction,
            private_key=self.from_private
        )
        # Send raw transaction
        self.w3.eth.send_raw_transaction(signed.rawTransaction)
        # get transaction _hash
        _hash = self.w3.toHex(self.w3.keccak(signed.rawTransaction))
        # Wait for transaction to finish or timeout
        start_time = datetime.now()
        while True:
            # We will wait for max 5minutes for this tx to respond, if it does not, we will re-try
            if datetime.now() - start_time > timedelta(minutes=TIMEOUT_MINS):
                success = False
                logging.info(f"Transaction {self}, timed out!")
                break
            try:
                recepit = self.w3.eth.get_transaction_receipt(_hash)
                if recepit["status"] == 1:
                    success = True
                else:
                    success = False
                break
            except exceptions.TransactionNotFound:
                # Sleep 10s while waiting
                sleep(10)
                logging.info(f"Waiting for transaction '{self}' to finish (Nonce:{nonce})...")

        if success:
            logging.info(f"Transaction {self} completed! hash: {_hash} - "
                        f"Explorer: https://explorer.roninchain.com/tx/{str(_hash)}")
            return _hash
        else:
            logging.info(f"Transaction {self} failed. Trying to augment gas price to unstuck it.")
            self.increase_gas_tx(nonce)

    def execute(self, gas_price=1, nonce=None):
        if self.token == 'ron':
            return self.execute_ron(gas_price, nonce)
        return self.execute_token(gas_price, nonce)

    def __str__(self):
        return f"Scatter of {self.token} from {self.from_acc.replace('0x', 'ronin:')}"


class TrezorScatter:
    def __init__(self, token, from_acc, client, bip_path, to_ronin_ammount_dict):
        self.w3 = Web3(
            Web3.HTTPProvider(
                RONIN_PROVIDER,
                request_kwargs={"headers": {"content-type": "application/json", "user-agent": USER_AGENT}}))
        self.token = token.lower()
        if self.token != 'ron':
            self.token_contract = self.w3.eth.contract(
                address=Web3.toChecksumAddress(TOKEN[self.token]),
                abi=APPROVE_ABI
            )
        self.from_acc = from_acc.replace("ronin:", "0x")
        self.client = client
        self.bip_path = parse_path(bip_path)
        self.contract = self.w3.eth.contract(
            address=Web3.toChecksumAddress(SCATTER_CONTRACT),
            abi=SCATTER_ABI
        )
        self.to_list = []
        self.amounts_list = []
        for k,v in to_ronin_ammount_dict.items():
            self.to_list.append(k.replace("ronin:", "0x"))
            self.amounts_list.append(v)
   
    def is_contract_accepted(self):        
        allowance = self.token_contract.functions.allowance(
            Web3.toChecksumAddress(self.from_acc),
            Web3.toChecksumAddress(SCATTER_CONTRACT)).call()
        if int(allowance) > sum(self.amounts_list):
            return True
        self.approve_contract()

    def approve_contract(self):
        nonce = get_nonce(self.from_acc)
        approve_tx = self.token_contract.functions.approve(
            Web3.toChecksumAddress(SCATTER_CONTRACT),
            115792089237316195423570985008687907853269984665640564039457584007913129639935
        ).buildTransaction({
            "gas": 1000000,
            "gasPrice": self.w3.toWei(1, "gwei"),
            "nonce": nonce
        })
        data = self.w3.toBytes(hexstr=approve_tx['data'])
        to = self.w3.toBytes(hexstr=TOKEN[self.token])
        sig = ethereum.sign_tx(
            self.client,
            n=self.bip_path,
            nonce=nonce,
            gas_price=self.w3.toWei(1, "gwei"),
            gas_limit=1000000,
            to=TOKEN[self.token],
            value=0,
            data=data,
            chain_id=2020
        )
        transaction = rlp.encode((nonce, self.w3.toWei(1, "gwei"), 1000000, to, 0, data) + sig)
        self.w3.eth.send_raw_transaction(transaction)
        approve_hash = self.w3.toHex(self.w3.keccak(transaction))
        approved = self.w3.eth.wait_for_transaction_receipt(approve_hash, timeout=240)
        if approved['status'] == 1:
            return True
        return False

    def increase_gas_tx(self, nonce):
        # check nonce is still available, do nothing if nonce is not available anymore
        if nonce != get_nonce(self.from_acc):
            return
        # Increase gas price to get tx unstuck
        if self.token == 'ron':
            self.execute_ron(1.01, nonce)
        self.execute_token(1.01, nonce)

    def execute_token(self, gas_price=1, nonce=None):
        # Check token is approved
        if not self.is_contract_accepted():
            logging.warning(f"Token {self.token} is not approved to use scatter, "
                            "you can re-try or manually accept it on "
                            "scatter website (https://scatter.roninchain.com/).")
            return

        # Check enough balance is present
        if not check_balance(self.from_acc, self.token) >= sum(self.amounts_list) and check_balance(self.from_acc, 'ron') >= 0.011:
            logging.warning("Not enough {TOKEN[self.token]} balance or not enough RON to pay for the tx")
            return
        
        # Get Nonce
        if nonce is None:
            nonce = get_nonce(self.from_acc)
        # Build transaction
        transaction = self.contract.functions.disperseTokenSimple(
            TOKEN[self.token],
            self.to_list,
            self.amounts_list
        ).buildTransaction({
            "chainId": 2020,
            "gas": 1000000,
            "gasPrice": self.w3.toWei(str(gas_price), "gwei"),
            "nonce": nonce
        })
        data = self.w3.toBytes(hexstr=transaction['data'])
        to = self.w3.toBytes(hexstr=SCATTER_CONTRACT)
        sig = ethereum.sign_tx(
            self.client,
            n=self.bip_path,
            nonce=nonce,
            gas_price=self.w3.toWei(str(gas_price), "gwei"),
            gas_limit=1000000,
            to=SCATTER_CONTRACT,
            value=0,
            data=data,
            chain_id=2020
        )
        transaction = rlp.encode((nonce, self.w3.toWei(str(gas_price), "gwei"), 1000000, to, 0, data) + sig)
        # Send raw transaction
        self.w3.eth.send_raw_transaction(transaction)
        _hash = self.w3.toHex(self.w3.keccak(transaction))
        # Wait for transaction to finish or timeout
        start_time = datetime.now()
        while True:
            # We will wait for max 5minutes for this tx to respond, if it does not, we will re-try
            if datetime.now() - start_time > timedelta(minutes=TIMEOUT_MINS):
                success = False
                logging.info(f"Transaction {self}, timed out!")
                break
            try:
                recepit = self.w3.eth.get_transaction_receipt(_hash)
                if recepit["status"] == 1:
                    success = True
                else:
                    success = False
                break
            except exceptions.TransactionNotFound:
                # Sleep 10s while waiting
                sleep(10)
                logging.info(f"Waiting for transaction '{self}' to finish (Nonce:{nonce})...")

        if success:
            logging.info(f"Transaction {self} completed! hash: {_hash} - "
                        f"Explorer: https://explorer.roninchain.com/tx/{str(_hash)}")
            return _hash
        else:
            logging.info(f"Transaction {self} failed. Trying to augment gas price to unstuck it.")
            self.increase_gas_tx(nonce)
    
    
    def execute_ron(self, gas_price=1, nonce=None):
        # Check enough balance is present
        if not check_balance(self.from_acc, 'ron') >= sum(self.amounts_list) + 0.011:
            logging.warning("Not enough RON balance to scatter and pay the tx.")
            return
                
        # Get Nonce
        if nonce is None:
            nonce = get_nonce(self.from_acc)
        # Build transaction
        transaction = self.contract.functions.disperseEther(
            self.to_list,
            self.amounts_list
        ).buildTransaction({
            "chainId": 2020,
            "gas": 1000000,
            "gasPrice": self.w3.toWei(str(gas_price), "gwei"),
            "nonce": nonce
        })
        data = self.w3.toBytes(hexstr=transaction['data'])
        to = self.w3.toBytes(hexstr=SCATTER_CONTRACT)
        sig = ethereum.sign_tx(
            self.client,
            n=self.bip_path,
            nonce=nonce,
            gas_price=self.w3.toWei(str(gas_price), "gwei"),
            gas_limit=1000000,
            to=SCATTER_CONTRACT,
            value=0,
            data=data,
            chain_id=2020
        )
        transaction = rlp.encode((nonce, self.w3.toWei(str(gas_price), "gwei"), 1000000, to, 0, data) + sig)
        # Send raw transaction
        self.w3.eth.send_raw_transaction(transaction)
        _hash = self.w3.toHex(self.w3.keccak(transaction))
        # Wait for transaction to finish or timeout
        start_time = datetime.now()
        while True:
            # We will wait for max 5minutes for this tx to respond, if it does not, we will re-try
            if datetime.now() - start_time > timedelta(minutes=TIMEOUT_MINS):
                success = False
                logging.info(f"Transaction {self}, timed out!")
                break
            try:
                recepit = self.w3.eth.get_transaction_receipt(_hash)
                if recepit["status"] == 1:
                    success = True
                else:
                    success = False
                break
            except exceptions.TransactionNotFound:
                # Sleep 10s while waiting
                sleep(10)
                logging.info(f"Waiting for transaction '{self}' to finish (Nonce:{nonce})...")

        if success:
            logging.info(f"Transaction {self} completed! hash: {_hash} - "
                        f"Explorer: https://explorer.roninchain.com/tx/{str(_hash)}")
            return _hash
        else:
            logging.info(f"Transaction {self} failed. Trying to augment gas price to unstuck it.")
            self.increase_gas_tx(nonce)

    def execute(self, gas_price=1, nonce=None):
        if self.token == 'ron':
            return self.execute_ron(gas_price, nonce)
        return self.execute_token(gas_price, nonce)

    def __str__(self):
        return f"Scatter of {self.token} from {self.from_acc.replace('0x', 'ronin:')}"



import logging

from eth_account.messages import encode_defunct
import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import RetryError
from web3 import Web3
from hexbytes import HexBytes
from trezorlib import ethereum
from trezorlib.tools import parse_path

from axie_utils.utils import USER_AGENT, RETRIES


class AxieGraphQL:
    def __init__(self, account, private_key, **kwargs):
        self.account = account.lower().replace("ronin:", "0x")
        self.private_key = private_key.lower()
        self.request = requests.Session()
        self.request.mount('https://', HTTPAdapter(max_retries=RETRIES))
        self.user_agent = USER_AGENT

    def create_random_msg(self):
        payload = {
            "operationName": "CreateRandomMessage",
            "variables": {},
            "query": "mutation CreateRandomMessage{createRandomMessage}"
        }
        url = "https://graphql-gateway.axieinfinity.com/graphql"
        try:
            response = self.request.post(url, json=payload)
        except RetryError as e:
            logging.critical(f"Error! Creating random msg! Error: {e}")
            return None
        if 200 <= response.status_code <= 299:
            try:
                return response.json()['data']['createRandomMessage']
            except KeyError:
                return None
        return None

    def get_jwt(self):
        msg = self.create_random_msg()
        if not msg:
            return None
        signed_msg = Web3().eth.account.sign_message(
            encode_defunct(text=msg),
            private_key=self.private_key
        )
        hex_msg = signed_msg['signature'].hex()
        payload = {
            "operationName": "CreateAccessTokenWithSignature",
            "variables": {
                "input": {
                    "mainnet": "ronin",
                    "owner": f"{self.account}",
                    "message": f"{msg}",
                    "signature": f"{hex_msg}"
                }
            },
            "query": "mutation CreateAccessTokenWithSignature($input: SignatureInput!)"
            "{createAccessTokenWithSignature(input: $input) "
            "{newAccount result accessToken __typename}}"
        }
        url = "https://graphql-gateway.axieinfinity.com/graphql"
        try:
            response = self.request.post(url, headers={"User-Agent": self.user_agent}, json=payload)
        except RetryError as e:
            logging.critical(f"Error! Getting JWT! Error: {e}")
            return None
        if 200 <= response.status_code <= 299:
            if (not response.json().get('data') or not response.json()['data'].get('createAccessTokenWithSignature') or
               not response.json()['data']['createAccessTokenWithSignature'].get('accessToken')):
                logging.critical("Could not retreive JWT, probably your private key for this account is wrong. "
                                 f"Account: {self.account.replace('0x','ronin:')} \n AccountName: {self.acc_name}")
                return None
            return response.json()['data']['createAccessTokenWithSignature']['accessToken']
        return None


class TrezorAxieGraphQL:
    def __init__(self, account, client, bip_path):
        self.account = account.lower().replace("ronin:", "0x")
        self.request = requests.Session()
        self.request.mount('https://', HTTPAdapter(max_retries=RETRIES))
        self.user_agent = USER_AGENT
        self.client = client
        self.bip_path = parse_path(bip_path)

    def create_random_msg(self):
        payload = {
            "operationName": "CreateRandomMessage",
            "variables": {},
            "query": "mutation CreateRandomMessage{createRandomMessage}"
        }
        url = "https://graphql-gateway.axieinfinity.com/graphql"
        try:
            response = self.request.post(url, json=payload)
        except RetryError as e:
            logging.critical(f"Error! Creating random msg! Error: {e}")
            return None
        if 200 <= response.status_code <= 299:
            try:
                return response.json()['data']['createRandomMessage']
            except KeyError:
                return None
        return None

    def get_jwt(self):
        msg = self.create_random_msg()
        if not msg:
            return None
        signed_msg = ethereum.sign_message(self.client, self.bip_path, msg)
        hex_msg = HexBytes(signed_msg.signature).hex()
        payload = {
            "operationName": "CreateAccessTokenWithSignature",
            "variables": {
                "input": {
                    "mainnet": "ronin",
                    "owner": f"{self.account}",
                    "message": f"{msg}",
                    "signature": f"{hex_msg}"
                }
            },
            "query": "mutation CreateAccessTokenWithSignature($input: SignatureInput!)"
            "{createAccessTokenWithSignature(input: $input) "
            "{newAccount result accessToken __typename}}"
        }
        url = "https://graphql-gateway.axieinfinity.com/graphql"
        try:
            response = self.request.post(url, headers={"User-Agent": self.user_agent}, json=payload)
        except RetryError as e:
            logging.critical(f"Error! Getting JWT! Error: {e}")
            return None
        if 200 <= response.status_code <= 299:
            if (not response.json().get('data') or not response.json()['data'].get('createAccessTokenWithSignature') or
               not response.json()['data']['createAccessTokenWithSignature'].get('accessToken')):
                logging.critical("Could not retreive JWT, probably your private key for this account is wrong. "
                                 f"Account: {self.account.replace('0x','ronin:')} \n AccountName: {self.acc_name}")
                return None
            return response.json()['data']['createAccessTokenWithSignature']['accessToken']
        return None



import logging
import rlp
from time import sleep
from datetime import datetime, timedelta

from web3 import Web3, exceptions
from trezorlib.tools import parse_path
from trezorlib import ethereum

from axie_utils.abis import AXIE_ABI
from axie_utils.utils import (
    get_nonce,
    RONIN_PROVIDER,
    AXIE_CONTRACT,
    TIMEOUT_MINS,
    USER_AGENT
)


class Breed:
    def __init__(self, sire_axie, matron_axie, address, private_key):
        self.w3 = Web3(
            Web3.HTTPProvider(
                RONIN_PROVIDER,
                request_kwargs={"headers": {"content-type": "application/json", "user-agent": USER_AGENT}}))
        self.sire_axie = sire_axie
        self.matron_axie = matron_axie
        self.address = address.replace("ronin:", "0x")
        self.private_key = private_key

    def execute(self):
        # Prepare transaction
        axie_contract = self.w3.eth.contract(
            address=Web3.toChecksumAddress(AXIE_CONTRACT),
            abi=AXIE_ABI
        )
        # Get Nonce
        nonce = get_nonce(self.address)
        # Build transaction
        transaction = axie_contract.functions.breedAxies(
            self.sire_axie,
            self.matron_axie
        ).buildTransaction({
            "chainId": 2020,
            "gas": 492874,
            "gasPrice": self.w3.toWei("1", "gwei"),
            "nonce": nonce
        })
        # Sign transaction
        signed = self.w3.eth.account.sign_transaction(
            transaction,
            private_key=self.private_key
        )
        # Send raw transaction
        self.w3.eth.send_raw_transaction(signed.rawTransaction)
        # get transaction _hash
        _hash = self.w3.toHex(self.w3.keccak(signed.rawTransaction))
        # Wait for transaction to finish or timeout
        logging.info("{self} about to start!")
        start_time = datetime.now()
        while True:
            # We will wait for max 10minutes for this tx to respond
            if datetime.now() - start_time > timedelta(minutes=TIMEOUT_MINS):
                success = False
                logging.info(f"Transaction {self}, timed out!")
                break
            try:
                recepit = self.w3.eth.get_transaction_receipt(_hash)
                if recepit["status"] == 1:
                    success = True
                else:
                    success = False
                break
            except exceptions.TransactionNotFound:
                # Sleep 10s while waiting
                sleep(10)
                logging.info(f"Waiting for transactions '{self}' to finish (Nonce: {nonce})...")

        if success:
            logging.info(f"{self} completed successfully")
            return _hash
        else:
            logging.info(f"{self} failed")

    def __str__(self):
        return (f"Breeding axie {self.sire_axie} with {self.matron_axie} in account "
                f"{self.address.replace('0x', 'ronin:')}")


class TrezorBreed:
    def __init__(self, sire_axie, matron_axie, address, client, bip_path):
        self.w3 = Web3(
            Web3.HTTPProvider(
                RONIN_PROVIDER,
                request_kwargs={"headers": {"content-type": "application/json", "user-agent": USER_AGENT}}))
        self.sire_axie = sire_axie
        self.matron_axie = matron_axie
        self.address = address.replace("ronin:", "0x")
        self.client = client
        self.bip_path = parse_path(bip_path)
        self.gwei = self.w3.toWei('1', 'gwei')
        self.gas = 250000

    def execute(self):
        # Prepare transaction
        axie_contract = self.w3.eth.contract(
            address=Web3.toChecksumAddress(AXIE_CONTRACT),
            abi=AXIE_ABI
        )
        # Get Nonce
        nonce = get_nonce(self.address)
        # Build transaction
        breed_tx = axie_contract.functions.breedAxies(
            self.sire_axie,
            self.matron_axie
        ).buildTransaction({
            "chainId": 2020,
            "gas": self.gas,
            "gasPrice": self.gwei,
            "nonce": nonce
        })
        data = self.w3.toBytes(hexstr=breed_tx['data'])
        to = self.w3.toBytes(hexstr=AXIE_CONTRACT)
        sig = ethereum.sign_tx(
            self.client,
            n=self.bip_path,
            nonce=nonce,
            gas_price=self.gwei,
            gas_limit=self.gas,
            to=AXIE_CONTRACT,
            value=0,
            data=data,
            chain_id=2020
        )
        transaction = rlp.encode((nonce, self.gwei, self.gas, to, 0, data) + sig)
        # Send raw transaction
        self.w3.eth.send_raw_transaction(transaction)
        # get transaction _hash
        _hash = self.w3.toHex(self.w3.keccak(transaction))
        # Wait for transaction to finish or timeout
        logging.info("{self} about to start!")
        start_time = datetime.now()
        while True:
            # We will wait for max 10minutes for this tx to respond
            if datetime.now() - start_time > timedelta(minutes=TIMEOUT_MINS):
                success = False
                logging.info(f"Transaction {self}, timed out!")
                break
            try:
                recepit = self.w3.eth.get_transaction_receipt(_hash)
                if recepit["status"] == 1:
                    success = True
                else:
                    success = False
                break
            except exceptions.TransactionNotFound:
                # Sleep 10s while waiting
                sleep(10)
                logging.info(f"Waiting for transactions '{self}' to finish (Nonce: {nonce})...")

        if success:
            logging.info(f"{self} completed successfully")
            return _hash
        else:
            logging.info(f"{self} failed")
            return

    def __str__(self):
        return (f"Breeding axie {self.sire_axie} with {self.matron_axie} in account "
                f"{self.address.replace('0x', 'ronin:')}")



import json
import logging
from datetime import datetime, timedelta

from web3 import Web3
import requests

from axie_utils.abis import AXIE_ABI
from axie_utils.utils import check_balance, RONIN_PROVIDER, AXIE_CONTRACT, USER_AGENT


class Axies:
    def __init__(self, account):
        self.w3 = Web3(
            Web3.HTTPProvider(
                RONIN_PROVIDER,
                request_kwargs={"headers": {"content-type": "application/json", "user-agent": USER_AGENT}}))
        self.acc = account.replace("ronin:", "0x")
        self.contract = self.w3.eth.contract(
            address=Web3.toChecksumAddress(AXIE_CONTRACT),
            abi=AXIE_ABI
        )
        self.now = datetime.now()

    def number_of_axies(self):
        return check_balance(self.acc, 'axies')

    def find_axies_to_morph(self):
        axie_list = self.get_axies()
        axies = []
        for axie in axie_list:
            morph_date, body_shape = self.get_morph_date_and_body(axie)
            if not morph_date and not body_shape:
                logging.info(f"Something went wrong getting info for Axie {axie}, skipping it")
            elif self.now >= morph_date and not body_shape:
                axies.append(axie)
            elif not body_shape:
                logging.info(f"Axie {axie} cannot be morphed until {morph_date}")
            else:
                logging.info(f"Axie {axie} is already an adult!")
        return axies

    def get_axies(self):
        num_axies = self.number_of_axies()
        axies = []
        for i in range(num_axies):
            axie = self.contract.functions.tokenOfOwnerByIndex(
                _owner=Web3.toChecksumAddress(self.acc),
                _index=i
            ).call()
            axies.append(axie)
        return axies

    @staticmethod
    def get_morph_date_and_body(axie_id):
        payload = {
            "operationName": "GetAxieDetail",
            "variables":
                {"axieId": axie_id},
            "query": "query GetAxieDetail($axieId: ID!) { axie(axieId: $axieId) "
            "{ ...AxieDetail __typename}} fragment AxieDetail on Axie "
            "{ id birthDate bodyShape __typename }"
        }
        url = "https://graphql-gateway.axieinfinity.com/graphql"
        response = requests.post(url, json=payload)
        try:
            json_response = response.json()
        except json.decoder.JSONDecodeError:
            logging.debug("Response contains no json info")
            return None, None

        if "data" in json_response and "axie" in json_response['data']:
            if 'bodyShape' in json_response['data']['axie'] and 'birthDate' in json_response['data']['axie']:
                # In case we want to check correctly morphed
                body_shape = json_response["data"]["axie"]["bodyShape"]
                birth_date = json_response["data"]["axie"]["birthDate"]
                morph_date = datetime.utcfromtimestamp(birth_date) + timedelta(days=5)
                return morph_date, body_shape

        return None, None

    @staticmethod
    def get_axie_details(axie_id):
        payload = {
            "operationName": "GetAxieDetail",
            "variables":
                {"axieId": axie_id},
            "query": "query GetAxieDetail($axieId: ID!) { axie(axieId: $axieId) "
            "{ ...AxieDetail }} fragment AxieDetail on Axie "
            "{ id class parts { ...AxiePart }} fragment AxiePart on AxiePart "
            "{ id name class type }"
        }
        url = "https://graphql-gateway.axieinfinity.com/graphql"
        response = requests.post(url, json=payload)
        try:
            json_response = response.json()
        except json.decoder.JSONDecodeError:
            logging.debug("Response contains no json info")
            return None
        if ("data" in json_response and 
            "axie" in json_response['data'] and 
            "parts" in json_response['data']['axie']):
            parts = {}
            for part in json_response['data']['axie']['parts']:
                parts[part['type'].lower()] = part['name'].lower()
            if "class" in json_response['data']['axie']:
                parts['class'] = json_response['data']['axie']['class'].lower()
            return parts
        return None



__version__ = '2.0.5'
__all__ = [
    'Axies',
    'AxieGraphQL',
    'Breed',
    'Claim',
    'CustomUI',
    'Morph',
    'Payment',
    'Scatter',
    'Transfer',
    'TrezorAxieGraphQL',
    'TrezorBreed',
    'TrezorClaim',
    'TrezorConfig',
    'TrezorMorph',
    'TrezorPayment',
    'TrezorScatter',
    'TrezorTransfer',
    'get_nonce',
    'get_lastclaim',
    'check_balance',
]

from axie_utils.axies import Axies
from axie_utils.breeding import Breed, TrezorBreed
from axie_utils.claims import Claim, TrezorClaim
from axie_utils.graphql import AxieGraphQL, TrezorAxieGraphQL
from axie_utils.morphing import Morph, TrezorMorph
from axie_utils.payments import Payment, TrezorPayment
from axie_utils.scatter import Scatter, TrezorScatter
from axie_utils.transfers import Transfer, TrezorTransfer
from axie_utils.utils import get_nonce, check_balance, CustomUI, TrezorConfig, get_lastclaim



AXIE_ABI = [{
    'constant': False,
    'inputs': [{'internalType': 'address', 'name': '_from',
               'type': 'address'}, {'internalType': 'address',
               'name': '_to', 'type': 'address'},
               {'internalType': 'uint256', 'name': '_tokenId',
               'type': 'uint256'}],
    'name': 'safeTransferFrom',
    'outputs': [{'internalType': 'bool', 'name': '_success',
                'type': 'bool'}],
    'payable': False,
    'stateMutability': 'nonpayable',
    'type': 'function',
    }, {
    'constant': False,
    'inputs': [{'internalType': 'uint256', 'name': '_sireId',
               'type': 'uint256'}, {'internalType': 'uint256',
               'name': '_matronId', 'type': 'uint256'}],
    'name': 'breedAxies',
    'outputs': [{'internalType': 'bool', 'name': '_success',
                'type': 'bool'}],
    'payable': False,
    'stateMutability': 'nonpayable',
    'type': 'function',
    }, {
    'constant': True,
    'inputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
    'name': 'balanceOf',
    'outputs': [{'internalType': 'uint256', 'name': '',
                'type': 'uint256'}],
    'payable': False,
    'stateMutability': 'view',
    'type': 'function',
    }, {
    'constant': True,
    'inputs': [{'internalType': 'address', 'name': '_owner',
               'type': 'address'}, {'internalType': 'uint256',
               'name': '_index', 'type': 'uint256'}],
    'name': 'tokenOfOwnerByIndex',
    'outputs': [{'internalType': 'uint256', 'name': '',
                'type': 'uint256'}],
    'payable': False,
    'stateMutability': 'view',
    'type': 'function',
    }]

SLP_ABI = [{
    'constant': False,
    'inputs': [{'internalType': 'address', 'name': '_owner',
               'type': 'address'}, {'internalType': 'uint256',
               'name': '_amount', 'type': 'uint256'},
               {'internalType': 'uint256', 'name': '_createdAt',
               'type': 'uint256'}, {'internalType': 'bytes',
               'name': '_signature', 'type': 'bytes'}],
    'name': 'checkpoint',
    'outputs': [{'internalType': 'uint256', 'name': '_balance',
                'type': 'uint256'}],
    'payable': False,
    'stateMutability': 'nonpayable',
    'type': 'function',
    }, {
    'constant': True,
    'inputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
    'name': 'balanceOf',
    'outputs': [{'internalType': 'uint256', 'name': '',
                'type': 'uint256'}],
    'payable': False,
    'stateMutability': 'view',
    'type': 'function',
    }, {
    'constant': False,
    'inputs': [{'internalType': 'address', 'name': '_to',
               'type': 'address'}, {'internalType': 'uint256',
               'name': '_value', 'type': 'uint256'}],
    'name': 'transfer',
    'outputs': [{'internalType': 'bool', 'name': '_success',
                'type': 'bool'}],
    'payable': False,
    'stateMutability': 'nonpayable',
    'type': 'function',
    }]

BALANCE_ABI = [{
    'constant': True,
    'inputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
    'name': 'balanceOf',
    'outputs': [{'internalType': 'uint256', 'name': '',
                'type': 'uint256'}],
    'payable': False,
    'stateMutability': 'view',
    'type': 'function',
    }]

APPROVE_ABI = [{
        'constant': False,
        'inputs': [
            {'name': '_spender', 'type':'address'},
            {'name': '_value', 'type': 'uint256'}],
        'name': 'approve',
        'outputs': [],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function'
    },{
        'constant': True,
        'inputs': [
            {'name': '_owner', 'type':'address'},
            {'name': '_spender', 'type': 'address'}],
        'name': 'allowance',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function'
    }]

SCATTER_ABI = [{
        'constant': False,
        'inputs': [
            {'name': 'token', 'type':'address'},
            {'name': 'recipients', 'type': 'address[]'},
            {'name': 'values', 'type': 'uint256[]'}],
        'name': 'disperseTokenSimple',
        'outputs': [],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function'
    }, {
        'constant': False,
        'inputs': [{'name': 'token', 'type': 'address'},{'name': 'recipients', 'type': 'address[]'},{'name': 'values', 'type': 'uint256[]'}],
        'name': 'disperseToken',
        'outputs': [],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function'
    }, {
        'constant': False,
        'inputs': [
        {'name': 'recipients', 'type': 'address[]'},{'name': 'values', 'type': 'uint256[]'}],
        'name': 'disperseEther',
        'outputs': [],
        'payable': True,
        'stateMutability': 'payable',
        'type': 'function'
    }]



import logging

from eth_account.messages import encode_defunct
from hexbytes import HexBytes
from trezorlib import ethereum
from requests.exceptions import RetryError
from web3 import Web3

from axie_utils.graphql import AxieGraphQL, TrezorAxieGraphQL


class Morph(AxieGraphQL):
    def __init__(self, axie, **kwargs):
        self.axie = axie
        super().__init__(**kwargs)

    def execute(self):
        jwt = self.get_jwt()
        msg = f"axie_id={self.axie}&owner={self.account}"
        signed_msg = Web3().eth.account.sign_message(encode_defunct(text=msg),
                                                     private_key=self.private_key)
        signature = signed_msg['signature'].hex()
        headers = {
            "User-Agent": self.user_agent,
            "authorization": f"Bearer {jwt}"
        }
        payload = {
            "operationName": "MorphAxie",
            "variables": {
                "axieId": f"{self.axie}",
                "owner": f"{self.account}",
                "signature": f"{signature}"
            },
            "query": "mutation MorphAxie($axieId: ID!, $owner: String!, $signature: String!) "
            "{morphAxie(axieId: $axieId, owner: $owner, signature: $signature)}"
        }
        url = 'https://graphql-gateway.axieinfinity.com/graphql'
        try:
            response = self.request.post(url, headers=headers, json=payload)
        except RetryError:
            logging.critical(f"Important! Axie {self.axie} in {self.account} is not ready to be morphed!")
            return

        if 200 <= response.status_code <= 299:
            if response.json().get('data') and response.json()['data'].get('morphAxie'):
                logging.info(f"Axie {self.axie} in {self.account} correctly morphed!")
                return
            else:
                logging.info(f"Something went wrong morphing axie {self.axie} in {self.account}")
                return
        else:
            logging.critical(f"Axie {self.axie} in {self.account} is not ready to be morphed!")
        return


class TrezorMorph(TrezorAxieGraphQL):
    def __init__(self, axie, **kwargs):
        self.axie = axie
        super().__init__(**kwargs)

    def execute(self):
        jwt = self.get_jwt()
        msg = f"axie_id={self.axie}&owner={self.account}"
        signed_msg = ethereum.sign_message(self.client, self.bip_path, msg)
        signature = HexBytes(signed_msg.signature).hex()
        headers = {
            "User-Agent": self.user_agent,
            "authorization": f"Bearer {jwt}"
        }
        payload = {
            "operationName": "MorphAxie",
            "variables": {
                "axieId": f"{self.axie}",
                "owner": f"{self.account}",
                "signature": f"{signature}"
            },
            "query": "mutation MorphAxie($axieId: ID!, $owner: String!, $signature: String!) "
            "{morphAxie(axieId: $axieId, owner: $owner, signature: $signature)}"
        }
        url = 'https://graphql-gateway.axieinfinity.com/graphql'
        try:
            response = self.request.post(url, headers=headers, json=payload)
        except RetryError:
            logging.critical(f"Axie {self.axie} in {self.account} is not ready to be morphed!")
            return

        if 200 <= response.status_code <= 299:
            if response.json().get('data') and response.json()['data'].get('morphAxie'):
                logging.info(f"Axie {self.axie} in {self.account} correctly morphed!")
            else:
                logging.info(f"Something went wrong morphing axie {self.axie} in {self.account}")
        else:
            logging.critical(f"Axie {self.axie} in {self.account} is not ready to be morphed!")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['axie_utils']

package_data = \
{'': ['*']}

install_requires = \
['trezor[hidapi]>=0.13.0,<0.14.0', 'web3>=5.25.0,<6.0.0']

setup_kwargs = {
    'name': 'axie-utils',
    'version': '2.0.5',
    'description': 'Library that provides the functionality you need to build your own axie Infinity python tools',
    'long_description': '# Axie Utils Library\n\nAim of this library is to contain all the actions one might want to do when building tools around the Axie Infinity videogame. It started with me building an automation tool and needing to build different solutions. Extracting this functionality allows for that easily.\n\n**NOTE: Only v1 of this library uses free tx, from v2 and onwards all transactions consume RON. That is due to now free tx being much more rare to have available.**\n\n\n# Installation\n\nInstall and update using pip:\n\n```\npip install -U axie-utils\n```\n\n# Simple Example\n\nThis example would send 100 SLP from `ronin:from_where_to_send_SLP` to `ronin:to_where_we_send_SLP`.\n\n``` python\nfrom axie_utils import Payment\n\np = Payment(\n    "Testing Account",\n    "ronin:from_where_to_send_SLP",\n    "0x:private_key_from_the_from_acc",\n    "ronin:to_where_we_send_SLP",\n    100)\n\np.execute()\n```\n\nThis example, shows how we would claim SLP from an account.\n\n``` python\nfrom axie_utils import Claim\n\nc = Claim(\n    "Testing Account",\n    "ronin:acc_to_claim",\n    "0x:private_key_from_acc_to_claim"\n)\nc.execute()\n\n```\n\n# Documentation\n\nFor furhter documentation, please visit this [link](https://ferranmarin.github.io/axie-utils-lib/).\n',
    'author': 'Ferran Marin',
    'author_email': 'ferran.marin.llobet@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://ferranmarin.github.io/axie-utils-lib/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import rlp
import logging
from datetime import datetime, timedelta
from time import sleep

from trezorlib import ethereum
from trezorlib.tools import parse_path
from web3 import Web3, exceptions

from axie_utils.abis import SLP_ABI
from axie_utils.utils import (
    get_nonce,
    SLP_CONTRACT,
    RONIN_PROVIDER,
    TIMEOUT_MINS,
    USER_AGENT
)


class Payment:
    def __init__(self, name, from_acc, from_private, to_acc, amount):
        self.w3 = Web3(
            Web3.HTTPProvider(
                RONIN_PROVIDER,
                request_kwargs={"headers": {"content-type": "application/json", "user-agent": USER_AGENT}}))
        self.name = name
        self.from_acc = from_acc.replace("ronin:", "0x")
        self.from_private = from_private
        self.to_acc = to_acc.replace("ronin:", "0x")
        self.amount = amount
        self.contract = self.w3.eth.contract(
            address=Web3.toChecksumAddress(SLP_CONTRACT),
            abi=SLP_ABI
        )

    def increase_gas_tx(self, nonce):
        # check nonce is still available, do nothing if nonce is not available anymore
        if nonce != get_nonce(self.from_acc):
            return
        # Increase gas price to get tx unstuck
        self.execute(1.01, nonce)

    def execute(self, gas_price=1, nonce=None):
        # Get Nonce
        if nonce is None:
            nonce = get_nonce(self.from_acc)
        # Build transaction
        transaction = self.contract.functions.transfer(
            Web3.toChecksumAddress(self.to_acc),
            self.amount
        ).buildTransaction({
            "chainId": 2020,
            "gas": 246437,
            "gasPrice": self.w3.toWei(str(gas_price), "gwei"),
            "nonce": nonce
        })
        # Sign Transaction
        signed = self.w3.eth.account.sign_transaction(
            transaction,
            private_key=self.from_private
        )
        # Send raw transaction
        self.w3.eth.send_raw_transaction(signed.rawTransaction)
        # get transaction _hash
        _hash = self.w3.toHex(self.w3.keccak(signed.rawTransaction))
        # Wait for transaction to finish or timeout
        start_time = datetime.now()
        while True:
            # We will wait for max 5minutes for this tx to respond, if it does not, we will re-try
            if datetime.now() - start_time > timedelta(minutes=TIMEOUT_MINS):
                success = False
                logging.info(f"Transaction {self}, timed out!")
                break
            try:
                recepit = self.w3.eth.get_transaction_receipt(_hash)
                if recepit["status"] == 1:
                    success = True
                else:
                    success = False
                break
            except exceptions.TransactionNotFound:
                # Sleep 10s while waiting
                sleep(10)
                logging.info(f"Waiting for transaction '{self}' to finish (Nonce:{nonce})...")

        if success:
            logging.info(f"Transaction {self} completed! _hash: {_hash} - "
                         f"Explorer: https://explorer.roninchain.com/tx/{str(_hash)}")
            return _hash
        else:
            logging.info(f"Transaction {self} failed. Trying to augment gas price to unstuck it.")
            self.increase_gas_tx(nonce)

    def __str__(self):
        return f"{self.name}({self.to_acc.replace('0x', 'ronin:')}) for the amount of {self.amount} SLP"


class TrezorPayment:
    def __init__(self, name, client, bip_path, from_acc, to_acc, amount):
        self.w3 = Web3(
            Web3.HTTPProvider(
                RONIN_PROVIDER,
                request_kwargs={"headers": {"content-type": "application/json", "user-agent": USER_AGENT}}))
        self.name = name
        self.from_acc = from_acc.replace("ronin:", "0x")
        self.to_acc = to_acc.replace("ronin:", "0x")
        self.amount = amount
        self.contract = self.w3.eth.contract(
            address=Web3.toChecksumAddress(SLP_CONTRACT),
            abi=SLP_ABI
        )
        self.client = client
        self.bip_path = parse_path(bip_path)
        self.gas = 250000

    def increase_gas_tx(self, nonce):
        # check nonce is still available, do nothing if nonce is not available anymore
        if nonce != get_nonce(self.from_acc):
            return
        # Increase gas price to get tx unstuck
        self.execute(1.01, nonce)

    def execute(self, gas_price=1, nonce=None):
        # Get Nonce
        if nonce is None:
            nonce = get_nonce(self.from_acc)
        # Build transaction
        send_tx = self.contract.functions.transfer(
            Web3.toChecksumAddress(self.to_acc),
            self.amount
        ).buildTransaction({
            "chainId": 2020,
            "gas": self.gas,
            "gasPrice": self.w3.toWei(str(gas_price), "gwei"),
            "nonce": nonce
        })
        data = self.w3.toBytes(hexstr=send_tx['data'])
        to = self.w3.toBytes(hexstr=SLP_CONTRACT)
        sig = ethereum.sign_tx(
            self.client,
            n=self.bip_path,
            nonce=nonce,
            gas_price=self.w3.toWei(str(gas_price), "gwei"),
            gas_limit=self.gas,
            to=SLP_CONTRACT,
            value=0,
            data=data,
            chain_id=2020
        )
        transaction = rlp.encode((nonce, self.w3.toWei(str(gas_price), 'gwei'), self.gas, to, 0, data) + sig)
        # Send raw transaction
        self.w3.eth.send_raw_transaction(transaction)
        _hash = self.w3.toHex(self.w3.keccak(transaction))
        # Wait for transaction to finish or timeout
        start_time = datetime.now()
        while True:
            # We will wait for max 5 minutes for this tx to respond, if it does not, we will re-try
            if datetime.now() - start_time > timedelta(minutes=TIMEOUT_MINS):
                success = False
                logging.info(f"Transaction {self}, timed out!")
                break
            try:
                recepit = self.w3.eth.get_transaction_receipt(_hash)
                if recepit["status"] == 1:
                    success = True
                else:
                    success = False
                break
            except exceptions.TransactionNotFound:
                # Sleep 10s while waiting
                sleep(10)
                logging.info(f"Waiting for transaction '{self}' to finish (Nonce:{nonce})...")

        if success:
            logging.info(f"Transaction {self} completed! _hash: {_hash} - "
                         f"Explorer: https://explorer.roninchain.com/tx/{str(_hash)}")
            return _hash
        else:
            logging.info(f"Transaction {self} failed. Trying to augment gas price to unstuck it.")
            self.increase_gas_tx(nonce)

    def __str__(self):
        return f"{self.name}({self.to_acc.replace('0x', 'ronin:')}) for the amount of {self.amount} SLP"



import asyncio
import rlp
import logging
from time import sleep
from datetime import datetime, timedelta, timezone

import requests
from requests.exceptions import RetryError
from web3 import Web3, exceptions
from trezorlib import ethereum

from axie_utils.abis import SLP_ABI
from axie_utils.utils import (
    check_balance,
    get_nonce,
    SLP_CONTRACT,
    RONIN_PROVIDER
)
from axie_utils.graphql import AxieGraphQL, TrezorAxieGraphQL


class Claim(AxieGraphQL):
    def __init__(self, acc_name, force, **kwargs):
        super().__init__(**kwargs)
        self.w3 = Web3(
            Web3.HTTPProvider(
                RONIN_PROVIDER,
                request_kwargs={"headers": {"content-type": "application/json", "user-agent": self.user_agent}}))
        self.slp_contract = self.w3.eth.contract(
            address=Web3.toChecksumAddress(SLP_CONTRACT),
            abi=SLP_ABI
        )
        self.acc_name = acc_name
        self.force = force
        self.request = requests.Session()

    def localize_date(self, date_utc):
        return date_utc.replace(tzinfo=timezone.utc).astimezone(tz=None)

    def humanize_date(self, date):
        local_date = self.localize_date(date)
        return local_date.strftime("%m/%d/%Y, %H:%M")
        
    def has_unclaimed_slp(self):
        url = f"https://game-api.skymavis.com/game-api/clients/{self.account}/items/1"
        try:
            response = self.request.get(url, headers={"User-Agent": self.user_agent})
        except RetryError:
            logging.critical(f"Failed to check if there is unclaimed SLP for acc {self.acc_name} "
                             f"({self.account.replace('0x','ronin:')})")
            return None
        if 200 <= response.status_code <= 299:
            data = response.json()
            last_claimed = datetime.utcfromtimestamp(data['last_claimed_item_at'])
            next_claim_date = last_claimed + timedelta(days=14)
            utcnow = datetime.utcnow()
            if utcnow < next_claim_date and not self.force:
                logging.critical(f"This account will be claimable again on {self.humanize_date(next_claim_date)}.")
                return None
            elif self.force:
                logging.info('Skipping check of dates, --force option was selected')
            wallet_total = check_balance(self.account)
            in_game_total = int(data['total'])
            if in_game_total > wallet_total:
                return in_game_total - wallet_total
        return None

    async def async_execute(self):
        unclaimed = self.has_unclaimed_slp()
        if not unclaimed:
            logging.info(f"Important: Account {self.acc_name} ({self.account.replace('0x', 'ronin:')}) "
                         "has no claimable SLP")
            return
        logging.info(f"Account {self.acc_name} ({self.account.replace('0x', 'ronin:')}) has "
                     f"{unclaimed} unclaimed SLP")
        jwt = self.get_jwt()
        if not jwt:
            logging.critical("Skipping claiming, we could not get the JWT for account "
                             f"{self.account.replace('0x', 'ronin:')}")
            return
        headers = {
            "User-Agent": self.user_agent,
            "authorization": f"Bearer {jwt}"
        }
        url = f"https://game-api.skymavis.com/game-api/clients/{self.account}/items/1/claim"
        try:
            response = self.request.post(url, headers=headers, json="")
        except RetryError as e:
            logging.critical(f"Error! Executing SLP claim API call for account {self.acc_name}"
                             f"({self.account.replace('0x', 'ronin:')}). Error {e}")
            return
        if 200 <= response.status_code <= 299:
            signature = response.json()["blockchain_related"].get("signature")
            if not signature or not signature["signature"]:
                logging.critical(f"Account {self.acc_name} ({self.account.replace('0x', 'ronin:')}) had no signature "
                                 "in blockchain_related")
                return
        else:
            logging.info(f"Claim for account {self.acc_name} ({self.account.replace('0x', 'ronin:')}) "
                         "had to be skipped")
            return
        nonce = get_nonce(self.account)
        # Build claim
        claim = self.slp_contract.functions.checkpoint(
            Web3.toChecksumAddress(self.account),
            signature['amount'],
            signature['timestamp'],
            signature['signature']
        ).buildTransaction({'gas': 492874, 'gasPrice': 1, 'nonce': nonce})
        # Sign claim
        signed_claim = self.w3.eth.account.sign_transaction(
            claim,
            private_key=self.private_key
        )
        # Send raw transaction
        self.w3.eth.send_raw_transaction(signed_claim.rawTransaction)
        # Get transaction hash
        hash = self.w3.toHex(self.w3.keccak(signed_claim.rawTransaction))
        # Wait for transaction to finish
        while True:
            try:
                recepit = self.w3.eth.get_transaction_receipt(hash)
                if recepit["status"] == 1:
                    success = True
                else:
                    success = False
                break
            except exceptions.TransactionNotFound:
                logging.debug(f"Waiting for claim for {self.acc_name} ({self.account.replace('0x', 'ronin:')}) to "
                              f"finish (Nonce:{nonce}) (Hash: {hash})...")
                # Sleep 5 seconds not to constantly send requests!
                await asyncio.sleep(5)
        if success:
            logging.info(f"SLP Claimed! New balance for account {self.acc_name} "
                         f"({self.account.replace('0x', 'ronin:')}) is: {check_balance(self.account)}")
        else:
            logging.info(f"Claim for account {self.acc_name} ({self.account.replace('0x', 'ronin:')}) "
                         "failed")

    def execute(self):
        unclaimed = self.has_unclaimed_slp()
        if not unclaimed:
            logging.info(f"Important: Account {self.acc_name} ({self.account.replace('0x', 'ronin:')}) "
                         "has no claimable SLP")
            return
        logging.info(f"Account {self.acc_name} ({self.account.replace('0x', 'ronin:')}) has "
                     f"{unclaimed} unclaimed SLP")
        jwt = self.get_jwt()
        if not jwt:
            logging.critical("Skipping claiming, we could not get the JWT for account "
                             f"{self.account.replace('0x', 'ronin:')}")
            return
        headers = {
            "User-Agent": self.user_agent,
            "authorization": f"Bearer {jwt}"
        }
        url = f"https://game-api.skymavis.com/game-api/clients/{self.account}/items/1/claim"
        try:
            response = self.request.post(url, headers=headers, json="")
        except RetryError as e:
            logging.critical(f"Error! Executing SLP claim API call for account {self.acc_name}"
                             f"({self.account.replace('0x', 'ronin:')}). Error {e}")
            return
        if 200 <= response.status_code <= 299:
            signature = response.json()["blockchain_related"].get("signature")
            if not signature or not signature["signature"]:
                logging.critical(f"Account {self.acc_name} ({self.account.replace('0x', 'ronin:')}) had no signature "
                                 "in blockchain_related")
                return
        else:
            logging.info(f"Claim for account {self.acc_name} ({self.account.replace('0x', 'ronin:')}) "
                         "had to be skipped")
            return
        nonce = get_nonce(self.account)
        # Build claim
        claim = self.slp_contract.functions.checkpoint(
            Web3.toChecksumAddress(self.account),
            signature['amount'],
            signature['timestamp'],
            signature['signature']
        ).buildTransaction({'gas': 492874, 'gasPrice': 1, 'nonce': nonce})
        # Sign claim
        signed_claim = self.w3.eth.account.sign_transaction(
            claim,
            private_key=self.private_key
        )
        # Send raw transaction
        self.w3.eth.send_raw_transaction(signed_claim.rawTransaction)
        # Get transaction hash
        hash = self.w3.toHex(self.w3.keccak(signed_claim.rawTransaction))
        # Wait for transaction to finish
        while True:
            try:
                recepit = self.w3.eth.get_transaction_receipt(hash)
                if recepit["status"] == 1:
                    success = True
                else:
                    success = False
                break
            except exceptions.TransactionNotFound:
                logging.debug(f"Waiting for claim for {self.acc_name} ({self.account.replace('0x', 'ronin:')}) to "
                              f"finish (Nonce:{nonce}) (Hash: {hash})...")
                # Sleep 5 seconds not to constantly send requests!
                sleep(5)
        if success:
            logging.info(f"SLP Claimed! New balance for account {self.acc_name} "
                         f"({self.account.replace('0x', 'ronin:')}) is: {check_balance(self.account)}")
        else:
            logging.info(f"Claim for account {self.acc_name} ({self.account.replace('0x', 'ronin:')}) "
                         "failed")


class TrezorClaim(TrezorAxieGraphQL):
    def __init__(self, acc_name, force, **kwargs):
        super().__init__(**kwargs)
        self.w3 = Web3(
            Web3.HTTPProvider(
                RONIN_PROVIDER,
                request_kwargs={"headers": {"content-type": "application/json", "user-agent": self.user_agent}}))
        self.slp_contract = self.w3.eth.contract(
            address=Web3.toChecksumAddress(SLP_CONTRACT),
            abi=SLP_ABI
        )
        self.acc_name = acc_name
        self.force = force
        self.request = requests.Session()
        self.gwei = self.w3.toWei('1', 'gwei')
        self.gas = 492874

    def localize_date(self, date_utc):
        return date_utc.replace(tzinfo=timezone.utc).astimezone(tz=None)

    def humanize_date(self, date):
        local_date = self.localize_date(date)
        return local_date.strftime("%m/%d/%Y, %H:%M")
        
        
    def has_unclaimed_slp(self):
        url = f"https://game-api.skymavis.com/game-api/clients/{self.account}/items/1"
        try:
            response = self.request.get(url, headers={"User-Agent": self.user_agent})
        except RetryError:
            logging.critical(f"Failed to check if there is unclaimed SLP for acc {self.acc_name} "
                             f"({self.account.replace('0x','ronin:')})")
            return None
        if 200 <= response.status_code <= 299:
            data = response.json()
            last_claimed = datetime.utcfromtimestamp(data['last_claimed_item_at'])
            next_claim_date = last_claimed + timedelta(days=14)
            utcnow = datetime.utcnow()
            if utcnow < next_claim_date and not self.force:
                logging.critical(f"This account will be claimable again on {self.humanize_date(next_claim_date)}.")
                return None
            elif self.force:
                logging.info('Skipping check of dates, --force option was selected')
            wallet_total = check_balance(self.account)
            in_game_total = int(data['total'])
            if in_game_total > wallet_total:
                return in_game_total - wallet_total
        return None

    async def async_execute(self):
        unclaimed = self.has_unclaimed_slp()
        if not unclaimed:
            logging.info(f"Account {self.acc_name} ({self.account.replace('0x', 'ronin:')}) "
                         "has no claimable SLP")
            return
        logging.info(f"Account {self.acc_name} ({self.account.replace('0x', 'ronin:')}) has "
                     f"{unclaimed} unclaimed SLP")
        jwt = self.get_jwt()
        if not jwt:
            logging.critical("Skipping claiming, we could not get the JWT for account "
                             f"{self.account.replace('0x', 'ronin:')}")
            return
        headers = {
            "User-Agent": self.user_agent,
            "authorization": f"Bearer {jwt}"
        }
        url = f"https://game-api.skymavis.com/game-api/clients/{self.account}/items/1/claim"
        try:
            response = self.request.post(url, headers=headers, json="")
        except RetryError as e:
            logging.critical(f"Error! Executing SLP claim API call for account {self.acc_name}"
                             f"({self.account.replace('0x', 'ronin:')}). Error {e}")
            return
        if 200 <= response.status_code <= 299:
            signature = response.json()["blockchain_related"].get("signature")
            if not signature or not signature["signature"]:
                logging.critical(f"Account {self.acc_name} ({self.account.replace('0x', 'ronin:')}) had no signature "
                                 "in blockchain_related")
                return
        else:
            logging.info(f"Claim for account {self.acc_name} ({self.account.replace('0x', 'ronin:')}) "
                         "had to be skipped")
            return
        nonce = get_nonce(self.account)
        # Build claim
        claim = self.slp_contract.functions.checkpoint(
            Web3.toChecksumAddress(self.account),
            signature['amount'],
            signature['timestamp'],
            signature['signature']
        ).buildTransaction({'gas': self.gas, 'gasPrice': 1, 'nonce': nonce})
        data = self.w3.toBytes(hexstr=claim['data'])
        to = self.w3.toBytes(hexstr=SLP_CONTRACT)
        sig = ethereum.sign_tx(
            self.client,
            n=self.bip_path,
            nonce=nonce,
            gas_price=self.gwei,
            gas_limit=self.gas,
            to=SLP_CONTRACT,
            value=0,
            data=data,
            chain_id=2020
        )
        transaction = rlp.encode((nonce, self.gwei, self.gas, to, 0, data) + sig)
        # Send raw transaction
        self.w3.eth.send_raw_transaction(transaction)
        hash = self.w3.toHex(self.w3.keccak(transaction))
        # Wait for transaction to finish
        while True:
            try:
                recepit = self.w3.eth.get_transaction_receipt(hash)
                if recepit["status"] == 1:
                    success = True
                else:
                    success = False
                break
            except exceptions.TransactionNotFound:
                logging.debug(f"Waiting for claim for {self.acc_name} ({self.account.replace('0x', 'ronin:')}) to "
                              f"finish (Nonce:{nonce}) (Hash: {hash})...")
                # Sleep 5 seconds not to constantly send requests!
                await asyncio.sleep(5)
        if success:
            logging.info(f"SLP Claimed! New balance for account {self.acc_name} "
                         f"({self.account.replace('0x', 'ronin:')}) is: {check_balance(self.account)}")
            return
        else:
            logging.info(f"Claim for account {self.acc_name} ({self.account.replace('0x', 'ronin:')}) "
                         "failed")
        return

    def execute(self):
        unclaimed = self.has_unclaimed_slp()
        if not unclaimed:
            logging.info(f"Account {self.acc_name} ({self.account.replace('0x', 'ronin:')}) "
                         "has no claimable SLP")
            return
        logging.info(f"Account {self.acc_name} ({self.account.replace('0x', 'ronin:')}) has "
                     f"{unclaimed} unclaimed SLP")
        jwt = self.get_jwt()
        if not jwt:
            logging.critical("Skipping claiming, we could not get the JWT for account "
                             f"{self.account.replace('0x', 'ronin:')}")
            return
        headers = {
            "User-Agent": self.user_agent,
            "authorization": f"Bearer {jwt}"
        }
        url = f"https://game-api.skymavis.com/game-api/clients/{self.account}/items/1/claim"
        try:
            response = self.request.post(url, headers=headers, json="")
        except RetryError as e:
            logging.critical(f"Error! Executing SLP claim API call for account {self.acc_name}"
                             f"({self.account.replace('0x', 'ronin:')}). Error {e}")
            return
        if 200 <= response.status_code <= 299:
            signature = response.json()["blockchain_related"].get("signature")
            if not signature or not signature["signature"]:
                logging.critical(f"Account {self.acc_name} ({self.account.replace('0x', 'ronin:')}) had no signature "
                                 "in blockchain_related")
                return
        else:
            logging.info(f"Claim for account {self.acc_name} ({self.account.replace('0x', 'ronin:')}) "
                         "had to be skipped")
            return
        nonce = get_nonce(self.account)
        # Build claim
        claim = self.slp_contract.functions.checkpoint(
            Web3.toChecksumAddress(self.account),
            signature['amount'],
            signature['timestamp'],
            signature['signature']
        ).buildTransaction({'gas': self.gas, 'gasPrice': 0, 'nonce': nonce})
        data = self.w3.toBytes(hexstr=claim['data'])
        to = self.w3.toBytes(hexstr=SLP_CONTRACT)
        sig = ethereum.sign_tx(
            self.client,
            n=self.bip_path,
            nonce=nonce,
            gas_price=self.gwei,
            gas_limit=self.gas,
            to=SLP_CONTRACT,
            value=0,
            data=data,
            chain_id=2020
        )
        transaction = rlp.encode((nonce, self.gwei, self.gas, to, 0, data) + sig)
        # Send raw transaction
        self.w3.eth.send_raw_transaction(transaction)
        hash = self.w3.toHex(self.w3.keccak(transaction))
        # Wait for transaction to finish
        while True:
            try:
                recepit = self.w3.eth.get_transaction_receipt(hash)
                if recepit["status"] == 1:
                    success = True
                else:
                    success = False
                break
            except exceptions.TransactionNotFound:
                logging.debug(f"Waiting for claim for {self.acc_name} ({self.account.replace('0x', 'ronin:')}) to "
                              f"finish (Nonce:{nonce}) (Hash: {hash})...")
                # Sleep 5 seconds not to constantly send requests!
                sleep(5)
        if success:
            logging.info(f"SLP Claimed! New balance for account {self.acc_name} "
                         f"({self.account.replace('0x', 'ronin:')}) is: {check_balance(self.account)}")
            return
        else:
            logging.info(f"Claim for account {self.acc_name} ({self.account.replace('0x', 'ronin:')}) "
                         "failed")
        return



import logging
from datetime import datetime
from json.decoder import JSONDecodeError

import requests
from requests.packages.urllib3.util.retry import Retry
from web3 import Web3
from trezorlib.ui import ClickUI
from trezorlib.client import get_default_client
from trezorlib.tools import parse_path
from trezorlib import ethereum

from axie_utils.abis import BALANCE_ABI

USER_AGENT = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/36.0.1944.0 Safari/537.36" # noqa
TIMEOUT_MINS = 5
AXIE_CONTRACT = "0x32950db2a7164ae833121501c797d79e7b79d74c"
AXS_CONTRACT = "0x97a9107c1793bc407d6f527b77e7fff4d812bece"
SLP_CONTRACT = "0xa8754b9fa15fc18bb59458815510e40a12cd2014"
WETH_CONTRACT = "0xc99a6a985ed2cac1ef41640596c5a5f9f4e19ef5"
USDC_CONTRACT = "0x0b7007c13325c48911f73a2dad5fa5dcbf808adc"
SCATTER_CONTRACT = "0x14978681c5f8ce2f6b66d1f1551b0ec67405574c"
RONIN_PROVIDER_FREE = "https://proxy.roninchain.com/free-gas-rpc"
RONIN_PROVIDER = "https://api.roninchain.com/rpc"
RETRIES = Retry(
    total=5,
    backoff_factor=2,
    status_forcelist=[500, 502, 503, 504],
    allowed_methods=frozenset(['GET', 'POST'])
)

TOKEN = {
    'slp': SLP_CONTRACT,
    'axs': AXS_CONTRACT,
    'axies': AXIE_CONTRACT,
    'weth': WETH_CONTRACT,
    'usdc': USDC_CONTRACT
}


def check_balance(account, token='slp'):
    w3 = Web3(
            Web3.HTTPProvider(
                RONIN_PROVIDER,
                request_kwargs={
                    "headers": {"content-type": "application/json",
                                "user-agent": USER_AGENT}}))
    if token.lower() in TOKEN:
        contract = TOKEN[token.lower()]
    elif token.lower() == "ron":
        return float(w3.eth.get_balance(Web3.toChecksumAddress(account.replace("ronin:", "0x"))) / 1000000000000000000)
    else:
        return 0
    ctr = w3.eth.contract(
        address=Web3.toChecksumAddress(contract),
        abi=BALANCE_ABI
    )
    balance = ctr.functions.balanceOf(
        Web3.toChecksumAddress(account.replace("ronin:", "0x"))
    ).call()
    if token == 'weth':
        return float(balance/1000000000000000000)
    return int(balance)


def get_nonce(account):
    w3 = Web3(
            Web3.HTTPProvider(
                RONIN_PROVIDER_FREE,
                request_kwargs={
                    "headers": {"content-type": "application/json",
                                "user-agent": USER_AGENT}}))
    nonce = w3.eth.get_transaction_count(
        Web3.toChecksumAddress(account.replace("ronin:", "0x"))
    )
    return nonce


def get_lastclaim(account):
    url = f'https://game-api.skymavis.com/game-api/clients/{account.replace("ronin:", "0x")}/items/1'
    try:
        r = requests.get(url)
        rjs = r.json()
        if rjs.get('last_claimed_item_at'):
            date = datetime.utcfromtimestamp(rjs['last_claimed_item_at'])
            return date
    except JSONDecodeError:
        logging.critical('Something went wrong getting last claim')

    return None


class CustomUI(ClickUI):
    def __init__(self, passphrase=None, *args, **kwargs):
        self.passphrase = passphrase
        super().__init__(*args, **kwargs)

    def get_passphrase(self, *args, **kwargs):
        return self.passphrase


class TrezorConfig:
    def __init__(self, accounts_number, passphrase=None):
        self.accounts_number = accounts_number
        self.passphrase = '' if not passphrase else passphrase

    def list_bip_paths(self):
        response = {}
        ui = CustomUI(passphrase=self.passphrase)
        for i in range(self.accounts_number):
            bip_path = f"m/44'/60'/0'/0/{i}"
            client = get_default_client(ui=ui)
            ronin = ethereum.get_address(client, parse_path(bip_path), True).lower().replace('0x', 'ronin:')
            response[ronin] = {"passphrase": self.passphrase, "bip_path": bip_path}

        return response
